import Vue, { ComponentOptions, PluginFunction, AsyncComponent } from "vue";

type Component = ComponentOptions<Vue> | typeof Vue | AsyncComponent;
type Dictionary<T> = { [ key:string ]:T };
type ErrorHandler = ( err:Error ) => void;

export type RouterMode = "hash" | "history" | "abstract";
export type RawLocation = string | Location;
export type RedirectOption = RawLocation | (( to:Route ) => RawLocation);
export type NavigationGuard<V extends Vue = Vue> = (
  to:Route,
  from:Route,
  next:( to?:RawLocation | false | (( vm:V ) => any) | void ) => void
) => any

export declare class VueRouter
{
  constructor( options?:RouterOptions );

  app:Vue;
  mode:RouterMode;
  currentRoute:Route;
  views : boolean;
  viewsConfig : ViewsConfig;

  beforeEach( guard:NavigationGuard ):Function;

  beforeResolve( guard:NavigationGuard ):Function;

  afterEach( hook:( to:Route, from:Route ) => any ):Function;

  push( location:RawLocation, onComplete?:Function, onAbort?:ErrorHandler ):void;

  replace( location:RawLocation, onComplete?:Function, onAbort?:ErrorHandler ):void;

  active( location:RawLocation, onComplete?:Function, onAbort?:ErrorHandler ):void;

  close( location:RawLocation, onComplete?:Function, onAbort?:ErrorHandler ):void;

  go( n:number ):void;

  back():void;

  forward():void;

  getMatchedComponents( to?:RawLocation | Route ):Component[];

  onReady( cb:Function, errorCb?:ErrorHandler ):void;

  onError( cb:ErrorHandler ):void;

  addRoutes( routes:RouteConfig[] ):void;

  resolve( to:RawLocation, current?:Route, append?:boolean ):{
    location:Location;
    route:Route;
    href:string;
    // backwards compat
    normalizedTo:Location;
    resolved:Route;
  };

  static install:PluginFunction<never>;
}

type Position = { x:number, y:number };
type PositionResult = Position | { selector:string, offset?:Position } | void;

export interface RouterOptions
{
  routes?:RouteConfig[];
  mode?:RouterMode;
  fallback?:boolean;
  views?:boolean | ViewsConfig;
  base?:string;
  linkActiveClass?:string;
  linkExactActiveClass?:string;
  parseQuery?:( query:string ) => Object;
  stringifyQuery?:( query:Object ) => string;
  scrollBehavior?:(
    to:Route,
    from:Route,
    savedPosition:Position | void
  ) => PositionResult | Promise<PositionResult>;
}

type patternTypes = (String | RegExp | Array<string>)[];

export interface ViewsConfig
{
  // 为 RouterViews 增加与 keep-alive 同样的优化参数逻辑
  // --------------------------------------------------------------------------
  // 只缓存包含规则的路由
  include:patternTypes,
  // 只缓存排除规则的路由
  exclude:patternTypes,
  // 缓存数量上限
  max:Number
}

type RoutePropsFunction = ( route:Route ) => Object;

export interface PathToRegexpOptions
{
  sensitive?:boolean;
  strict?:boolean;
  end?:boolean;
}

export interface RouteConfig
{
  path:string;
  name?:string;
  component?:Component;
  components?:Dictionary<Component>;
  redirect?:RedirectOption;
  alias?:string | string[];
  children?:RouteConfig[];
  meta?:any;
  beforeEnter?:NavigationGuard;
  props?:boolean | Object | RoutePropsFunction;
  caseSensitive?:boolean;
  pathToRegexpOptions?:PathToRegexpOptions;
}

export interface RouteRecord
{
  path:string;
  regex:RegExp;
  components:Dictionary<Component>;
  instances:Dictionary<Vue>;
  name?:string;
  parent?:RouteRecord;
  redirect?:RedirectOption;
  matchAs?:string;
  meta:any;
  beforeEnter?:(
    route:Route,
    redirect:( location:RawLocation ) => void,
    next:() => void
  ) => any;
  props:boolean | Object | RoutePropsFunction | Dictionary<boolean | Object | RoutePropsFunction>;
}

export interface Location
{
  // 用于 router.active 和 router.close 方法来操作多视图数据
  key?:string;

  name?:string;
  path?:string;
  hash?:string;
  query?:Dictionary<string | (string | null)[] | null | undefined>;
  params?:Dictionary<string>;
  append?:boolean;
  replace?:boolean;
}

export interface Route
{
  // 用于 beforeEach、afterEach 等方法来返回由 ViewsModel 生成的对路由的 key 值
  // 因存在同一个路由打开多个视图的情况，单纯依赖 path 无法识别唯一性，必须依赖 key 值区分
  key?:string;

  path:string;
  name?:string;
  hash:string;
  query:Dictionary<string | (string | null)[]>;
  params:Dictionary<string>;
  fullPath:string;
  matched:RouteRecord[];
  redirectedFrom?:string;
  meta?:any;
}
